import os
import subprocess
from dataclasses import dataclass, field
from typing import Optional, List, Dict, Any, Sequence
import yaml

@dataclass
class STaRArguments:
    star_config: str = "videostar/star_config.yaml"
    star_path: str = "data/star_json"
    train_script_path: str = "videollava/train/train_mem.py"
    data_gen_path: str = "videostar/data_generation.py"
    train_args_path: str = "videostar/videollava_args.yaml"
    star_data_path: Optional[List[str]] = field(default=None, metadata={"help": "Path to the training data."})
    aux_data_path: Optional[List[str]] = field(default=None, metadata={"help": "Path to the auxiliary data."})
    model_path: str = "LanguageBind/Video-LLaVA-7B"


def parse_yaml_args(yaml_path: str) -> Dict[str, Any]:
    with open(yaml_path, 'r') as file:
        return yaml.safe_load(file)

def run_training_script(script_path: str, args: List[str]):
    command = ["deepspeed", script_path] + args
    subprocess.run(command, env={**os.environ})

def run_generation(script_path: str, args: List[str]):
    command = ["accelerate", "launch", script_path] + args
    subprocess.run(command, env={**os.environ}, check=True)

def update_train_args_for_cycle(train_args: Dict[str, Any], cycle: int):
    train_args['output_dir'] = f"./checkpoints/Video-LLaVA-7B-STaR{cycle}"
    train_args['data_path'] = [
        f"data/star_json/STaR{cycle}/kinetics700_tune_.json",
        f"data/star_json/STaR{cycle}/finediving_tune_.json",
        f"data/star_json/STaR{cycle}/starb_tune_.json",
        f"data/star_json/STaR{cycle}/somsom_tune_.json",
        f"data/star_json/STaR{cycle}/moma_tune_.json",
        f"data/star_json/STaR{cycle}/egoexo4d_tune_.json",
        f"data/ft_json/videochatgpt_filtered_tune_.json",
        f"data/ft_json/llava_image_tune_.json",
        f"data/ft_json/nlp_tune.json"
    ]

def star_train():
    # Load STaRArguments from a fixed path
    star_args = STaRArguments()

    # Load training arguments from YAML file
    train_args = parse_yaml_args(star_args.train_args_path)
    star_config = parse_yaml_args(star_args.star_config)
    train_script_path = star_args.train_script_path

    model_path = star_args.model_path
    for cycle, gen_target in enumerate(star_config['cycles']):
        gen_args  = ['generation_target', int(gen_target), 'dataset_jsons', star_config['star_datasets']]
        gen_args  = ['generation_target', int(gen_target)]
        gen_args  = []
        run_generation(star_args.data_gen_path, gen_args)
        
        # Update the train arguments for the current cycle
        train_args = update_train_args_for_cycle(train_args, cycle)

        model_path = train_args.output_dir

        # Convert updated train arguments to list of strings
        train_args_list = []
        for key, value in train_args.items():
            if isinstance(value, list):
                for item in value:
                    train_args_list.append(f"--{key}")
                    train_args_list.append(str(item))
            else:
                train_args_list.append(f"--{key}")
                train_args_list.append(str(value))

        run_training_script(star_args.train_script_path, train_args_list)

if __name__ == "__main__":
    star_train()
